/***********************************************************************/
/*                                                                     */
/* Slurp.hpp   : Header file for mmap utility class                    */
/*                                                                     */
/*               Copyright (c) Yak! 2005 All rights reserved.          */
/*                                                                     */
/*   $URL: svn://localhost/afxspdir/trunk/Slurp.hpp $                  */
/*  $Date: 2005-11-12 17:45:03 +0900 (Sat, 12 Nov 2005) $              */
/*   $Rev: 49 $                                                        */
/*                                                                     */
/***********************************************************************/

#ifndef SLURP_HPP
#define SLURP_HPP

#include <algorithm>
#include <iterator>

#include <windef.h>

namespace yak {
	namespace utils {
		namespace windows {

			class SlurpBase
			{
			public:
				explicit SlurpBase(LPSTR pszFilename, DWORD dwOffset = 0);
				explicit SlurpBase(HANDLE hFile_, DWORD dwOffset = 0);
				virtual ~SlurpBase();
				template<typename T>
				T* getPointer(size_t offset) {
					return reinterpret_cast<T*>(pbStart + offset);
				}
				template<typename T>
				const T* getPointer(size_t offset) const {
					return reinterpret_cast<const T*>(pbStart + offset);
				}
				template<typename T>
				T* begin() { return getPointer<T>(0); }
				template<typename T>
				T* begin() const { return getPointer<T>(0); }
				template<typename T>
				T* end() { return getPointer<T>(dwSize); }
				template<typename T>
				T* end() const { return getPointer<T>(dwSize); }
				DWORD getSize() const { return dwSize; }
				void swap(SlurpBase &r) {
					std::swap(hFile, r.hFile);
					std::swap(hMap, r.hMap);
					std::swap(pbActualStart, r.pbActualStart);
					std::swap(pbStart, r.pbStart);
					std::swap(dwSize, r.dwSize);
				}
			private:
				SlurpBase& operator=(const SlurpBase&); // Prohibit copy
				SlurpBase(const SlurpBase&);            // Prohibit copy
				void init(HANDLE hFile_, DWORD dwOffset);

				HANDLE hFile;
				HANDLE hMap;
				LPBYTE pbActualStart;
				LPBYTE pbStart;
				DWORD dwSize;
			};

			template<typename T>
			class Slurp : public SlurpBase
			{
			public:
				// types as STL container
				typedef T value_type;
				typedef T* iterator;
				typedef const T* const_iterator;
				typedef std::reverse_iterator<iterator> reverse_iterator;
				typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
				typedef T& reference;
				typedef const T& const_reference;
				typedef std::size_t size_type;
				typedef std::ptrdiff_t difference_type;

				// constructor/copy/destruct
				explicit Slurp(LPSTR pszFilename, DWORD dwOffset = 0):SlurpBase(pszFilename, dwOffset) {}
				explicit Slurp(HANDLE hFile, DWORD dwOffset = 0):SlurpBase(hFile, dwOffset) {}
				virtual ~Slurp() {}

				// iterator support
				iterator begin() { return SlurpBase::begin<T>(); }
				const_iterator begin() const { return SlurpBase::begin<T>(); }
				iterator end() { return SlurpBase::end<T>(); }
				const_iterator end() const { return SlurpBase::end<T>(); }

				// reverse iterator support
				reverse_iterator rbegin() { return std::reverse_iterator<iterator>(end()); }
				const_reverse_iterator rbegin() const { return std::reverse_iterator<const_iterator>(end()); };
				reverse_iterator rend()  { return std::reverse_iterator<iterator>(begin()); };
				const_reverse_iterator rend() const { return std::reverse_iterator<const_iterator>(begin()); };

				// capacity
				size_type size() { return getSize()/sizeof(T); }
				bool empty() { return size() != 0; }
				size_type max_size() { return size(); }

				// element access
				reference operator[](size_type index) { return *getPointer<T>(sizeof(T)*index); }
				const_reference operator[](size_type index) const  { return *getPointer<T>(sizeof(T)*index); };
				reference at(size_type index)  { return *getPointer<T>(sizeof(T)*index); } // TODO: should throw
				const_reference at(size_type index) const  { return *getPointer<T>(sizeof(T)*index); }  // TODO: should throw
				reference front() { return *begin(); }
				const_reference front() const { return *begin(); }
				reference back() { return *--end(); } // TODO: should throw if empty
				const_reference back() const { return *--end(); }

			private:
				Slurp& operator=(const Slurp&); // Prohibit copy
				Slurp(const Slurp&);            // Prohibit copy
			};

		} // namespace windows
	} // namespace utils
} // namespace yak

#endif
